////////////////////////////////////////////////////////////////////////
/*! 
 *  \file   test.cpp
 *  \brief  TiffIO plugin self-test
 *  \author JD Gascuel
 * 
 * Copyright JD Gascuel, 2005, 2010.
 * 
 * Jean-Dominique.Gascuel@imag.fr
 *
 * This software is a computer program whose purpose is to add TIFF image
 * read/write capabilities to the Qt library, hence to any Qt application.
 * 
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 *
 * TiffIO is based on the libtiff 3.9.4, with the following copyright :
 *      Copyright (c) 1988-1997 Sam Leffler
 *      Copyright (c) 1991-1997 Silicon Graphics, Inc.
 * See the libtiff-3.9.4/ subdirectory.
 */
////////////////////////////////////////////////////////////////////////
// File History:
//
// $Id: test.cpp 13 2010-06-25 16:31:07Z gascuel $
//
// 2005-02-01 : Creation.
// 2005-06-20 : added JPEG test (one image).

#include <stdio.h>      // fprintf()
#include <stdlib.h>     // ?
#include <math.h>       // fabs()

#include <qglobal.h>

#if QT_VERSION < 0x040000 //------------ Qt 3.x ------------------------------
#   include <qapplication.h> 
#   include <qimage.h>
#   include <qdir.h>
#   include <qimage.h>
#else                     //------------ Qt 4.x ------------------------------
#   include <QApplication>
#   include <QDir>
#   include <QImage>
#   include <QImageWriter>
#   include <QtCore/qplugin.h>
#   define  latin1() toLatin1().constData()
#   define  hasAlphaBuffer hasAlphaChannel
#   define  cleanDirPath cleanPath
#endif

//----------------------------------------------------------------------------
//
// KLUDGE for Windows: if any library is compiled in release (eg. Qt3 or Qt4),
// then the execution depends of the release C/C++ runtime library too:
// (copied from crtdefs.h)
#if defined(_DEBUG) && defined(WIN32)
// CUDA (both 1.0 and 1.1) and Qt are compiled against the non-debug 
// runtime C library.
// When this application is compiled in DEBUG mode, the manifest don't
// transitively reference the files.
// Hence, add some magic (from crtdefs.h) to ask the linker to know this...
    #if _M_AMD64
        #pragma comment(linker,"/manifestdependency:\"type='win32' " \
            "name='" __LIBRARIES_ASSEMBLY_NAME_PREFIX ".CRT' "       \
            "version='" _CRT_ASSEMBLY_VERSION "' "                   \
            "processorArchitecture='amd64' "                         \
            "publicKeyToken='" _VC_ASSEMBLY_PUBLICKEYTOKEN "'\"")
    #else
        #pragma comment(linker,"/manifestdependency:\"type='win32' " \
            "name='" __LIBRARIES_ASSEMBLY_NAME_PREFIX ".CRT' "       \
            "version='" _CRT_ASSEMBLY_VERSION "' "                   \
            "processorArchitecture='x86' "                           \
            "publicKeyToken='" _VC_ASSEMBLY_PUBLICKEYTOKEN "'\"")

    #endif
#endif

//----------------------------------------------------------------------------
//
// Fake handler to make sure an error occured while testing unsupported files.
//
static int qFatals = 0, qWarnings = 0, qDebugs = 0;
static char errorMsg[120];

static void
fakeMsgHandler(QtMsgType type, const char *m)
{
    // fprintf(stdout, "Trapped message = '%s'\n", msg);
    switch( type ) {
#if QT_VERSION >= 0x040000
        case QtSystemMsg: // aka. QtCriticalMsg
#endif
        case QtFatalMsg:   qFatals++;   break;
        case QtWarningMsg: qWarnings++; break;
        case QtDebugMsg:   qDebugs++;   break;
    }
    strncpy(errorMsg, m, sizeof errorMsg);
    errorMsg[(sizeof errorMsg)-1] = 0;
}

//-------------------------------------------------------------------------
// Make sure handler is loaded.
//
// On Qt4 : we have a way to ask for the plugin.
// For Qt3, this is done by hand inside TiffIO_Qt3.cpp.

#ifdef QT_STATICPLUGIN
Q_IMPORT_PLUGIN(TIFFIOPlugin)
#endif

inline int MAX(int a, int b) { return a<b ? b : a; }

static bool pixelsDiffers(const QImage ref, const QImage& img, int q)
{
    int traceErrors = 5;    // Dumps only the first 5 errors.

    // Maximum allowed error: Just rounding if quality==99, else a %
    // of the possible range (0..255) :
    int dmax = (q==99) ? 1 : (((100-q) * 255)+99)/100;

    for(int j=0; j<img.height(); ++j)
        for(int i=0; i<img.width(); ++i)
        {
            QRgb  a = ref.pixel(i,j), 
                  b = img.pixel(i,j);
            int d = abs( qRed  (a) - qRed  (b));
            d = MAX(d, abs( qGreen(a) - qGreen(b)));
            d = MAX(d, abs( qBlue (a) - qBlue (b)));
            d = MAX(d, abs( qAlpha(a) - qAlpha(b)));

            if( d > dmax )
            {
                fprintf(stderr, "\tPixel(%2d,%2d) : %08X != %08X\n",
                    i, j, a, b);
                if( --traceErrors < 0 )
                {
                    fprintf(stderr, "\t...\n");
                    return true;
                }
            }
        }

    return traceErrors < 5; // Anything printed ?
}
//----------------------------------------------------------------------------
//
//
int main(int argc, char *argv[])
{
    QApplication app(argc, argv, false);

    // Make sure that no installed plugins will be taken
    // from the $QTDIR/plugins architecture.
#if defined(Q_STATIC_PLUGIN) || defined(QT_STATICPLUGIN)
    app.setLibraryPaths( QStringList() );
#endif

    // Check any error messages generated by Qt or LibTiff
    qInstallMsgHandler(fakeMsgHandler);

#ifdef JPEG_SUPPORT
    fprintf(stdout, "    jpeg-in-tiff supported\n");
#else
    fprintf(stdout, "*** NO JPEG SUPPORT ***\n");
#endif
#ifdef ZIP_SUPPORT
    fprintf(stdout, "    zip compression support\n");
#else
    fprintf(stdout, "*** NO ZLIB SUPPORT ***\n");
#endif
#ifndef LZW_SUPPORT
    fprintf(stdout, "*** NO LZW SUPPORT ***\n");
#elif defined(LZW_SUPPORT) && defined(NO_LZW_ENCODE)
    fprintf(stdout, "    lzw DECOMPRESSION-ONLY support\n");
#else
    fprintf(stdout, "    lzw compression support\n");
#endif

#ifdef PACKBITS_SUPPORT
    fprintf(stdout, "    packbits (aka rle) supported\n");
#endif
#ifdef PIXARLOG_SUPPORT
    fprintf(stdout, "    pixarlog supported\n");
#endif
#ifdef PIXARFILM_SUPPORT
    fprintf(stdout, "    pixarfilm supported\n");
#endif

    //-------------------------------------------------------------------------
    //-------------------------------------------------------------------------
    // A bunch of various combination of formats, all created with PhotoShop.
    struct ReadExamples {
        const char *name;                // File path, inside Examples/
        int w, h, depth;                 // Image size.
        unsigned int ll, ul, lr, ur;     // Values of the 4 corner pixels, ll = (0,0)
    } readExamples[] = {
        // NAME                             W    H    d     LL (redish)      UL (yellow)       LR (blue)        UR (greenish)

        // Check byte orders :
        {"rgb8-none-mac.tif",               100, 100, 32,   qRgb(252,0,126),  qRgb(252,251,0),  qRgb(1,0,251),   qRgb(2,251,126)},
        {"rgb8-none-pc.tif",                100, 100, 32,   qRgb(252,0,126),  qRgb(252,251,0),  qRgb(1,0,251),   qRgb(2,251,126)},
        {"rgb8-none-planar.tif",            100, 100, 32,   qRgb(252,0,126),  qRgb(252,251,0),  qRgb(1,0,251),   qRgb(2,251,126)},
        {"rgb8-none-planar-mac.tif",        100, 100, 32,   qRgb(252,0,126),  qRgb(252,251,0),  qRgb(1,0,251),   qRgb(2,251,126)},

        {"rgb16-none-pc.tif",               100, 100, 32,   qRgb(252,0,126),  qRgb(252,251,0),  qRgb(1,0,252),   qRgb(1,251,126)},
        {"rgb16-none-mac.tif",              100, 100, 32,   qRgb(252,0,126),  qRgb(252,251,0),  qRgb(1,0,252),   qRgb(1,251,126)},
        {"rgb16-none-planar.tif",           100, 100, 32,   qRgb(252,0,126),  qRgb(252,251,0),  qRgb(1,0,252),   qRgb(1,251,126)},
        {"rgb16-none-planar-mac.tif",       100, 100, 32,   qRgb(252,0,126),  qRgb(252,251,0),  qRgb(1,0,252),   qRgb(1,251,126)},

        // Check bit orders :
        {"rgb8-lsb2msb.tif",                100, 100, 32,   qRgb(254,0,93),  qRgb(254,254,0),  qRgb(1,0,254),   qRgb(1,254,95)},
        {"rgb8-msb2lsb.tif",                100, 100, 32,   qRgb(254,0,93),  qRgb(254,254,0),  qRgb(1,0,254),   qRgb(1,254,95)},

        // Check Strip vs tiles :
        {"rgb8-separate.tif",               100, 100, 32,   qRgb(254,0,93),  qRgb(254,254,0),  qRgb(1,0,254),   qRgb(1,254,95)},
        {"rgb8-tiled.tif",                  100, 100, 32,   qRgb(254,0,93),  qRgb(254,254,0),  qRgb(1,0,254),   qRgb(1,254,95)},

        // Check alpha plane :
        // (note we have pre-multiplied image, hence pixel values are slightly different).
        {"rgb8-none-alpha.tif",             100, 100, 32,   qRgba(252,0,126,255), qRgba(252,251,1,233), qRgba(1,0,251,220), qRgba(1,251,126,204)},
        {"rgb8-none-alpha-planar.tif",      100, 100, 32,   qRgba(252,0,126,255), qRgba(252,251,1,233), qRgba(1,0,251,220), qRgba(1,251,126,204)},
        {"rgb8-none-alpha-planar-mac.tif",  100, 100, 32,   qRgba(252,0,126,255), qRgba(252,251,1,233), qRgba(1,0,251,220), qRgba(1,251,126,204)},
        {"rgb16-none-alpha.tif",            100, 100, 32,   qRgba(252,0,126,255), qRgba(252,251,0,234), qRgba(1,0,252,220), qRgba(1,251,126,204)},
        {"rgb16-none-alpha-mac.tif",        100, 100, 32,   qRgba(252,0,126,255), qRgba(252,251,0,234), qRgba(1,0,252,220), qRgba(1,251,126,204)},

        // Color spaces :
        {"cmyk8-none.tif",                  100, 100, 32,   qRgb(255,20,232), qRgb(229,255,22), qRgb(30,58,255), qRgb(94,255,69)},

        {"gray8-none.tif",                  100, 100, 8,    qRgb(96,96,96),     qRgb(239,239,239),  qRgb(39,39,39), qRgb(202,202,202)},
        {"gray16-none-pc.tif",              100, 100, 8,    qRgb(96,96,96),     qRgb(239,239,239),  qRgb(39,39,39), qRgb(202,202,202)},
        {"gray16-none-mac.tif",             100, 100, 8,    qRgb(96,96,96),     qRgb(239,239,239),  qRgb(39,39,39), qRgb(202,202,202)},

        {"color16-none.tif",                100, 100, 8,    qRgb(226,26,128),   qRgb(226,228,24),   qRgb(28,21,230),    qRgb(27,221,130)},
        {"color64-none.tif",                100, 100, 8,    qRgb(234,12,132),   qRgb(244,241,11),   qRgb(14,10,242),    qRgb(19,239,123)},
        {"color256-none.tif",               100, 100, 8,    qRgb(249,8,122),    qRgb(247,247,6),    qRgb(8,4,248),      qRgb(8,247,125)},

        {"bitmap-none.tif",                 100, 100, 1,    qRgb(255,255,255), qRgb(255,255,255),    qRgb(0,0,0), qRgb(255,255,255)},
        {"bitmap-lsb.tif",                  100, 100, 1,    qRgb(255,255,255), qRgb(255,255,255),    qRgb(255,255,255), qRgb(255,255,255)},

#ifdef CCITT_SUPPORT
        {"2480x3496-faxg4.tif",             2480,3496, 1,   qRgb(255,255,255), qRgb(255,255,255), qRgb(255,255,255), qRgb(255,255,255)},
        {"bitmap-fax=3.tif",                100, 100, 1,    qRgb(255,255,255), qRgb(255,255,255),    qRgb(0,0,0), qRgb(255,255,255)},
        {"bitmap-fax=4.tif",                100, 100, 1,    qRgb(255,255,255), qRgb(255,255,255),    qRgb(0,0,0), qRgb(255,255,255)},
#endif

#ifdef LZW_SUPPORT
        {"cmyk8-lzw.tif",                   100, 100, 32,   qRgb(255,3,139), qRgb(239,255,10), qRgb(29,58,255), qRgb(101,255,4)},
        {"lab8-lzw.tif",                    100, 100, 32,   qRgb(244,0,85),  qRgb(255,233,0),  qRgb(73,86,147), qRgb(135,191,53)},
        {"lab16-lzw.tif",                   100, 100, 32,   qRgb(255,0,113), qRgb(255,249,0),  qRgb(0,0,223),   qRgb(90,250,110)},
        {"rgb8-lzw-mac.tif",                100, 100, 32,   qRgb(254,0,93),  qRgb(254,254,0),  qRgb(1,0,254),   qRgb(1,254,95)},
        {"rgb8-lzw-pc.tif",                 100, 100, 32,   qRgb(254,0,93),  qRgb(254,254,0),  qRgb(1,0,254),   qRgb(1,254,95)},
        {"rgb16-lzw-mac.tif",               100, 100, 32,   qRgb(252,0,126), qRgb(252,251,0),  qRgb(1,0,252),   qRgb(1,251,126)},
        {"rgb16Lzw-layersZip.tif",          100, 100, 32,   qRgb(252,0,126),  qRgb(252,251,0),  qRgb(1,0,252),   qRgb(1,251,126)},
        {"color64-lzw-mac.tif",             100, 100, 8,    qRgb(247,18,84), qRgb(241,241,4),  qRgb(11,11,243), qRgb(20,246,85)},
        {"color64-lzw-pc.tif",              100, 100, 8,    qRgb(247,18,84), qRgb(241,241,4),  qRgb(11,11,243), qRgb(20,246,85)},
        {"gray8-lzw-mac.tif",               100, 100, 8,    qRgb(131,131,131), qRgb(247,247,247), qRgb(71,71,71), qRgb(219,219,219)},
        {"gray16-lzw-pc.tif",               100, 100, 8,    qRgb(130,130,130), qRgb(247,247,247), qRgb(71,71,71), qRgb(220,220,220)},
        {"bitmap-lzw-pc.tif",               100, 100, 1,    qRgb(255,255,255), qRgb(255,255,255),    qRgb(0,0,0), qRgb(255,255,255)},
#endif

#ifdef ZIP_SUPPORT
        {"rgb8-zip-mac.tif",                100, 100, 32,   qRgb(254,0,93),  qRgb(254,254,0),  qRgb(1,0,254),   qRgb(1,254,95)},
        {"rgb8-zip-pc.tif",                 100, 100, 32,   qRgb(254,0,93),  qRgb(254,254,0),  qRgb(1,0,254),   qRgb(1,254,95)},
        {"rgb16-zip-pyrm.tif",              100, 100, 32,   qRgb(252,0,126),  qRgb(252,251,0),  qRgb(1,0,252),   qRgb(1,251,126)},
        {"gray8-zip-pc.tif",                100, 100, 8,    qRgb(131,131,131), qRgb(247,247,247), qRgb(71,71,71), qRgb(219,219,219)},
        {"gray16-zip-mac.tif",              100, 100, 8,    qRgb(130,130,130), qRgb(247,247,247), qRgb(71,71,71), qRgb(220,220,220)},
        {"bitmap-zip-pc.tif",               100, 100, 1,    qRgb(255,255,255), qRgb(255,255,255),    qRgb(0,0,0), qRgb(255,255,255)},
#endif

#ifdef JPEG_SUPPORT
        {"gray8-jpeg.tif",                  100, 100, 8,    qRgb(130,130,130), qRgb(246,246,246), qRgb(70,70,70), qRgb(219,219,219)},
        {"rgb8-jpeg-YCrCb.tif",             100, 100, 32,   qRgb(255,0,94),  qRgb(255,254,3),  qRgb(2,1,253),   qRgb(3,252,96)},
        {"rgb8-jpeg-RGB.tif",               100, 100, 32,   qRgb(254,0,93),  qRgb(254,254,0),  qRgb(2,0,254),   qRgb(2,254,96)},
#endif

#ifdef PIXARLOG_SUPPORT
        {"rgb8-pixarlog.tif",               100, 100, 32,   qRgb(252,0,126),      qRgb(252,251,0),      qRgb(1,0,251),      qRgb(2,251,126)},
        {"rgb8-pixarlog-alpha.tif",         100, 100, 32,   qRgba(252,0,126,255), qRgba(252,251,1,233), qRgba(1,0,251,220), qRgba(1,251,126,204)},
#endif

        {0, 0, 0, 0, 0, 0, 0, 0}
    };

    int error = 0;
    for(ReadExamples *t = readExamples; t->name; t++)
    {
        QString name = QString("Examples/%1").arg(t->name);

        if( ! QFileInfo(name).exists() )
        {
            fprintf(stderr, "%d: missing file %s\n",
                ++error, t->name);
            continue;
        }

        // ---------------------------------------------------------- Read test
        QImage img( name, "TIFF" );

        if( img.isNull() )
            fprintf(stderr, "%d: Cannot load image %s\n", ++error, t->name);
        else if( img.width() != t->w || img.height() != t->h || img.depth() != t->depth )
            fprintf(stderr, "%d: Wrong header for %s. %dx%dx%d should be %dx%dx%d\n",
                ++error, t->name,
                img.width(), img.height(), img.depth(),
                t->w, t->h, t->depth);
        else
        {
            if( img.hasAlphaBuffer() != (strstr(t->name, "alpha")!=0) )
                fprintf(stderr, "%d: %s have wrong alpha channel configuration\n", 
                    ++error, t->name);

            if( img.pixel(0,0) != t->ll )
                fprintf(stderr, "%d: %s wrong pixel value.\n  ll %d,%d,%d,%d should be %d,%d,%d,%d\n",
                    ++error, t->name,
                    qRed(img.pixel(0,0)), qGreen(img.pixel(0,0)), qBlue(img.pixel(0,0)), qAlpha(img.pixel(0,0)),
                    qRed(t->ll),          qGreen(t->ll),          qBlue(t->ll),          qAlpha(t->ll));

            if( img.pixel(0,t->h-1) != t->ul)
                fprintf(stderr, "%d: %s wrong pixel value.\n  ul %d,%d,%d,%d should be %d,%d,%d,%d\n",
                    ++error, t->name,
                    qRed(img.pixel(0,t->h-1)), qGreen(img.pixel(0,t->h-1)), qBlue(img.pixel(0,t->h-1)), qAlpha(img.pixel(0,t->h-1)),
                    qRed(t->ul),               qGreen(t->ul),               qBlue(t->ul),               qAlpha(t->ul));

            if( img.pixel(t->w-1,0) != t->lr)
                fprintf(stderr, "%d: %s wrong pixel value.\n  lr %d,%d,%d,%d should be %d,%d,%d,%d\n",
                    ++error, t->name,
                    qRed(img.pixel(t->w-1,0)), qGreen(img.pixel(t->w-1,0)), qBlue(img.pixel(t->w-1,0)), qAlpha(img.pixel(t->w-1,0)),
                    qRed(t->lr),               qGreen(t->lr),               qBlue(t->lr),               qAlpha(t->lr));

            if( img.pixel(t->w-1,t->h-1) != t->ur )
                fprintf(stderr, "%d: %s wrong pixel value.\n  ur %d,%d,%d,%d should be %d,%d,%d,%d\n",
                    ++error, t->name,
                    qRed(img.pixel(t->w-1,t->h-1)), qGreen(img.pixel(t->w-1,t->h-1)), qBlue(img.pixel(t->w-1,t->h-1)), qAlpha(img.pixel(t->w-1,t->h-1)),
                    qRed(t->ur),                    qGreen(t->ur),                    qBlue(t->ur),                    qAlpha(t->ur));
        }

        if( qFatals || qWarnings || qDebugs )
        {
            fprintf(stderr, "%d: Spurious %s READ error: %s\n",
                ++error, t->name, errorMsg);
            qFatals = qWarnings = qDebugs = 0;
        }
    }

    //----------------------------------------------------------------------------
    // Write test.
    // Check various combination of what is possible to save.
    //
    // Problem: How to test the compression hints have been indeed understood ?
    // Solution (partial) : check the compressed file size. This should be constant
    //                      from one platform to the other, and should not change through
    //                      time (eg. a regression test indeed).
    //
    // Caveat: ZIP and JPEG are external libs. If implementation changes, they can
    //         become more (or less) efficient... and break the regression test.
    //
    // Note: Qt only support RGBA, RGB, Gray8 and colors format. Hence no needs to
    //       read other source examples...

    struct WriteExamples {
        const char* name;       // Source image to start with.
        int         quality;    // Optional quality hint (<100 means don't check pixels)
        const char* parameters; // Optional save hints.

        int         depth;      // Depth of the re-read image.
        bool        hasAlpha;   // Re-read image have an alpha channel ?
        int         numColors;  // Re-read image have a color table ? size ?
        uint        checkSize;  // Hintented saved file size +/-5%.
    } writeExamples[] =
    {
    //  Source name             Quality Save parameters      D  alpha  #col fileSize
        {"rgb8-none-alpha.tif", 100,    "default",          32, true,   0,  0},
        {"rgb8-none-pc.tif",    100,    "default",          32, false,  0,  0},
        {"gray8-none.tif",      100,    "default",           8, false, 256, 0},
        {"color256-none.tif",   100,    "default",           8, false, 256, 0},
        {"color64-none.tif",    100,    "default",           8, false, 256, 0},
        {"color16-none.tif",    100,    "default",           8, false, 256, 0},
        {"bitmap-none.tif",     100,    "default",           1, false,   2, 0},

        {"rgb8-none-alpha.tif", 100,    "uncompressed",     32, true,    0, 40230},
        {"rgb8-none-alpha.tif", 100,    "none",             32, true,    0, 40230},
        {"rgb8-none-pc.tif",    100,    "uncompressed",     32, false,   0, 30210},
        {"gray8-none.tif",      100,    "none",              8, false, 256, 10134},
        {"color256-none.tif",   100,    "none",              8, false, 256, 11682},
        {"color64-none.tif",    100,    "uncompressed",      8, false, 256, 11682},
        {"color16-none.tif",    100,    "none",              8, false, 256, 11682},
        {"bitmap-none.tif",     100,    "uncompressed",      1, false,   2,  1446},

        {"rgb8-none-alpha.tif", 100,    "none bps=16",      32, true,    0, 80278},
        {"rgb8-none-alpha.tif", 100,    "none bps=12",      32, true,    0, 60278},
        {"rgb8-none-alpha.tif", 100,    "none bps=10",      32, true,    0, 50230},
        {"rgb8-none-alpha.tif",  93,    "none bps=7",       32, true,    0, 35230},
        {"rgb8-none-alpha.tif",  87,    "none bps=6",       32, true,    0, 30214},
        {"rgb8-none-alpha.tif",  75,    "none bps=5",       32, true,    0, 25206},
        {"rgb8-none-alpha.tif",  50,    "none bps=4",       32, true,    0, 20198},
        {"rgb8-none-alpha.tif",  25,    "none bps=3",       32, true,    0, 15198},
        {"rgb8-none-alpha.tif",  12,    "none bps=2",       32, true,    0, 10190},
        {"rgb8-none-alpha.tif",   6,    "none bps=1",       32, true,    0,  5174},

        {"rgb8-none-pc.tif",    100,    "none bps=16",      32, false,   0, 60274},
        {"rgb8-none-pc.tif",    100,    "none bps=12",      32, false,   0, 45226},
        {"rgb8-none-pc.tif",    100,    "none bps=10",      32, false,   0, 37726},
        {"rgb8-none-pc.tif",     93,    "none bps=7",       32, false,   0, 26510},
        {"rgb8-none-pc.tif",     87,    "none bps=6",       32, false,   0, 22702},
        {"rgb8-none-pc.tif",     75,    "none bps=5",       32, false,   0, 18994},
        {"rgb8-none-pc.tif",     50,    "none bps=4",       32, false,   0, 15194},
        {"rgb8-none-pc.tif",     25,    "none bps=3",       32, false,   0, 11486},
        {"rgb8-none-pc.tif",     12,    "none bps=2",       32, false,   0,  7670},
        {"rgb8-none-pc.tif",      6,    "none bps=1",       32, false,   0,  3970},

        {"gray8-none.tif",      100,    "none bps=16",       8, false, 256, 20134},
        {"gray8-none.tif",      100,    "none bps=12",       8, false, 256, 15134},
        {"gray8-none.tif",      100,    "none bps=10",       8, false, 256, 12634},
        {"gray8-none.tif",       93,    "none bps=7",        8, false, 128,  8934},
        {"gray8-none.tif",       87,    "none bps=6",        8, false,  64,  7634},
        {"gray8-none.tif",       75,    "none bps=5",        8, false,  32,  6434},
        {"gray8-none.tif",       50,    "none bps=4",        8, false,  16,  5134},
        {"gray8-none.tif",       25,    "none bps=3",        8, false,   8,  3934},
        {"gray8-none.tif",       12,    "none bps=2",        8, false,   4,  2634},
        {"gray8-none.tif",        6,    "none bps=1",        1, false,   2,  1434},

        {"color256-none.tif",   100,    "none bps=8",        8, false, 256,  11682},

        {"color64-none.tif",    100,    "none bps=7",        8, false, 128,  9714},
        {"color64-none.tif",    100,    "none bps=6",        8, false,  64,  8030},

        {"color16-none.tif",    100,    "none bps=7",        8, false, 128,  9714},
        {"color16-none.tif",    100,    "none bps=6",        8, false,  64,  8030},
        {"color16-none.tif",    100,    "none bps=5",        8, false,  32,  6638},
        {"color16-none.tif",    100,    "none bps=4",        8, false,  16,  5242},

#ifdef PACKBITS_SUPPORT
        // Note that RLE performs poorly (no compression) when there is gradients,
        // because there is no predictor, hence the encoded data is never repeated.
        {"rgb8-none-alpha.tif", 100,    "rle",              32, true,   0, 40584},
        {"rgb8-none-alpha.tif", 100,    "packbits",         32, true,   0, 40584},
        {"rgb8-none-pc.tif",    100,    "rle",              32, false,  0, 29704},
        {"gray8-none.tif",      100,    "packbits",          8, false, 256, 8992},
        {"color256-none.tif",   100,    "rle",               8, false, 256, 5336},
        {"color64-none.tif",    100,    "rle",               8, false, 256, 3378},
        {"color64-none.tif",    100,    "rle bps=6",         8, false,  64, 8006},
        {"color16-none.tif",    100,    "rle",               8, false, 256, 2608},
        {"color16-none.tif",    100,    "rle bps=4",         8, false,  16, 1484},
        {"bitmap-none.tif",     100,    "rle",               1, false,   2, 1538},  // 1446 uncompressed...
#endif

#ifdef CCITT_SUPPORT
        // Exists only for bitmap files...
        {"bitmap-none.tif",     100,    "fax",               1, false,   2, 3114},  // 1446 uncompressed...
        {"bitmap-none.tif",     100,    "fax=3",             1, false,   2, 3080},  // 1446 uncompressed...
        {"bitmap-none.tif",     100,    "fax=4",             1, false,   2, 3114},  // 1446 uncompressed...
#endif

#if defined(LZW_SUPPORT) && !defined(NO_LZW_ENCODE)
        {"rgb8-none-alpha.tif", 100,    "lzw",              32, true,   0, 43162},
        {"rgb8-none-alpha.tif",  50,    "lzw bps=4",        32, true,   0,  9042},
        {"rgb8-none-pc.tif",    100,    "lzw",              32, false,  0, 35562},
        {"rgb8-none-pc.tif",     50,    "lzw bps=4",        32, false,  0,  8350},
        {"gray8-none.tif",      100,    "lzw",               8, false, 256, 6058},
        {"gray8-none.tif",       50,    "lzw bps=4",         8, false,  16, 1442},
        {"color256-none.tif",   100,    "lzw",               8, false, 256, 5964},
        {"color64-none.tif",    100,    "lzw",               8, false, 256, 3738},
        {"color64-none.tif",    100,    "lzw bps=6",         8, false,  64, 2670},
        {"color16-none.tif",    100,    "lzw",               8, false, 256, 2718},
        {"color16-none.tif",    100,    "lzw bps=4",         8, false,  16, 1168},
        {"bitmap-none.tif",     100,    "lzw",               1, false,   2, 1634},  // 1446 uncompressed...
#endif

#ifdef ZIP_SUPPORT
        {"rgb8-none-alpha.tif", 100,    "zip=9",            32, true,   0, 33640},
        {"rgb8-none-alpha.tif",  50,    "zip=9 bps=4",      32, true,   0,  3776},
        {"rgb8-none-pc.tif",    100,    "zip=8",            32, false,  0, 24696},
        {"rgb8-none-pc.tif",     50,    "zip=8 bps=4",      32, false,  0,  3916},
        {"gray8-none.tif",      100,    "zip=7",             8, false, 256, 4670},
        {"color256-none.tif",   100,    "zip=5",             8, false, 256, 3690},
        {"color64-none.tif",    100,    "zip",               8, false, 256, 2430},     // Defaut : zip=6
        {"color64-none.tif",    100,    "zip bps=6",         8, false,  64, 1552},
        {"color16-none.tif",    100,    "zip",               8, false, 256, 2014},
        {"color16-none.tif",    100,    "zip bps=4",         8, false,  16,  600},
        {"bitmap-none.tif",     100,    "zip",               1, false,   2, 1394},
#endif

#ifdef JPEG_SUPPORT
        // Beware: this is a lossy codec. Asking for quality==100 don't garanty
        // image distance to be 0% ! In practice, we get:
        // Q=100 : error max ~ 2%
        // Q= 75 : error max ~ 14%
        {"rgb8-none-alpha.tif", 97,     "jpeg=100",         32, true,   0,  19511},
        {"rgb8-none-pc.tif",    97,     "jpeg=100",         32, false,  0,  13925},
        {"gray8-none.tif",      85,     "jpeg=75",           8, false, 256, 1667},
#endif

#ifdef PIXARLOG_SUPPORT
        // Test is not working on MacX : a different compression ration achieved
        // from the same sources ???
        // {"rgb8-none-alpha.tif",100,     "pixarlog=1",       32, true,   0, 26860},
        // {"rgb8-none-pc.tif",   100,     "pixarlog=1",       32, false,  0, 18848},
        {"rgb8-none-alpha.tif",100,     "pixarlog",         32, true,   0, 23602},
        {"rgb8-none-alpha.tif",100,     "pixarlog=6",       32, true,   0, 23602},
        {"rgb8-none-alpha.tif",100,     "pixarlog=9",       32, true,   0, 23458},
        {"rgb8-none-pc.tif",   100,     "pixarlog",         32, false,  0, 16644},
        {"rgb8-none-pc.tif",   100,     "pixarlog=6",       32, false,  0, 16644},
        {"rgb8-none-pc.tif",   100,     "pixarlog=9",       32, false,  0, 16622},
#endif

#ifdef PIXARFILM_SUPPORT
        {"rgb8-none-alpha.tif", 99,     "pixarfilm",        32, true,   0, 22094},
        {"rgb8-none-pc.tif",    99,     "pixarfilm",        32, false,  0, 16496},
#endif
        {0, 0, 0, 0, false, 0, 0}
    };

    // Get access to a writable directory.
    QString tmp = QDir::cleanDirPath(getenv("TMP"));
#if defined(WIN32)
    if( tmp.isEmpty() ) tmp = QDir::cleanDirPath(getenv("TEMP"));
    if( tmp.isEmpty() ) tmp = "C:/Temp";
#elif defined(UNIX)
    if( tmp.isEmpty() ) tmp = "/tmp";
#else
#error Should provide for a writable temporary directory
#endif

    printf("Using temp dir = %s\n", tmp.latin1());
    QDir::current().mkdir(tmp);   // In case it does not exists...
    tmp += "/test-";

    for(WriteExamples* t = writeExamples; t->name; ++t)
    {
        QImage img( QString("Examples/") + t->name );
        if( img.isNull() )
        {
            fprintf(stderr, "%d: cannot load image %s for write test\n",
                ++error, t->name);
        }
        else
        {
            QString name = tmp
                         + QString(t->name).section('.', 0, -2)
                         + "-" + t->parameters 
                         + ".tif";

#if QT_VERSION < 0x040000
            QImageIO writer(name, "TIFF");
            writer.setQuality(t->quality);
            writer.setParameters(t->parameters);
            writer.setImage(img);

            if( !  writer.write() )
            {
                fprintf(stderr, "%d: Unable to save to '%s' : %s\n", ++error,
                    name.latin1(),
                    errorMsg);
                qFatals = qWarnings = qDebugs = 0;
                continue;
            }
#else
            QFile file(name);
            if( !file.open(QIODevice::WriteOnly | QIODevice::Truncate) )
            {
                fprintf(stderr, "%d: unable to open %s for write test\n",
                    ++error, name.latin1());
                continue;
            }

            QImageWriter writer(&file, "TIFF");
            writer.setQuality(t->quality);
            writer.setText("parameters", t->parameters);

            if( ! writer.write(img) )
            {
                if( errorMsg[0]==0 )
                {
                    strncpy(errorMsg, 
                            writer.errorString().latin1(), 
                            sizeof errorMsg);
                    errorMsg[(sizeof errorMsg)-1] = 0;
                }

                fprintf(stderr, "%d: Unable to save to '%s': %s\n", ++error,
                    name.latin1(),
                    errorMsg);
                qFatals = qWarnings = qDebugs = 0;
                continue;
            }
#endif

            if( qFatals || qWarnings || qDebugs )
            {
                fprintf(stderr, "%d: Spurious %s WRITE error: %s\n",
                    ++error, t->parameters, errorMsg);
                qFatals = qWarnings = qDebugs = 0;
            }

            // Checking file size seems a fairly good test that we have all color
            // channels, bits, and we did pass the compression ratio requested to
            // libtiff. But, slight differences in the compression lib used
            // (eg. lzw in 3.6 against 3.8) gives slightly differente results, hence
            // the 1% allowance below...
            // NOTE: Beware of a uint/int bug under MacOSX : cast everyone to float first !
            QFileInfo fi(name);
            if( t->checkSize
             && fabsf((float(fi.size()) - float(t->checkSize))/float(t->checkSize)) > 0.01f )
            {
                fprintf(stderr, "%d: %s have wrong saved file size. %d should be %d (+/- 1%%)\n",
                    ++error,
                    name.section('/',-1).latin1(),
                    (int)fi.size(),
                    t->checkSize);
            }

            QImage reload(name);
            if( qFatals || qWarnings || qDebugs )
            {
                fprintf(stderr, "%d: Spurious %s REREAD error: %s\n",
                    ++error, t->parameters, errorMsg);
                qFatals = qWarnings = qDebugs = 0;
            }

            if( img.width() != reload.width() || img.height() != reload.height() )
                fprintf(stderr, "%d: %s %s write test failed, images size differ: %dx%d should be %dx%d\n",
                    ++error, t->name, t->parameters,
                    reload.width(), reload.height(), img.width(), img.height());
            else if( t->depth != reload.depth() || t->numColors != reload.numColors() )
                fprintf(stderr, "%d: %s %s write test failed, images mode differ: %d/%d should be %d/%d\n",
                    ++error, t->name, t->parameters,
                    reload.depth(), reload.numColors(), t->depth, t->numColors);
            else if( t->hasAlpha != reload.hasAlphaBuffer() )
                fprintf(stderr, "%d: %s %s write test failed, reloaded image should %shave alpha channel.\n",
                    ++error, t->name, t->parameters,
                    t->hasAlpha ? "" : "not ");
            else if( pixelsDiffers(img, reload, t->quality) )
                fprintf(stderr, "%d: %s %s write test failed, images pixels differ\n", 
                    ++error, t->name, t->parameters);

#ifndef DEBUG
            QDir::current().remove(name);
#endif
        }
    }

    //----------------------------------------------------------------------------
    // A bunch of images that are not supported, but should not crash the library,
    // Just generates [trapable] qFatal()/qWarning().
    static const char *unsuportedExamples[] = {
        "cmyk16-none.tif",                      // Limited to 8bits.
        "rgb16-none-alpha-planar.tif",          // Alpha limited to contig planes.
        "rgb16-none-alpha-planar-mac.tif",      // Idem.
        "gray8-none-alpha.tif",                 // 2 channels not supported.
        "gray16-none-alpha.tif",                // Idem.

/****** Should test those configurations don't work either, because there is not
        enough bits to store the pixel indexes.
        {"color64-none.tif",     50,    "none bps=5",        8, false,  32,  6638},
        {"color64-none.tif",     40,    "none bps=4",        8, false,  16,  5242},
        {"color64-none.tif",     30,    "none bps=3",        8, false,   8,  3994},
        {"color64-none.tif",     20,    "none bps=2",        8, false,   4,  2670},
        {"color64-none.tif",     10,    "none bps=1",        1, false,   2,  1458},
        {"color16-none.tif",     30,    "none bps=3",        8, false,   8,  3994},
        {"color16-none.tif",     20,    "none bps=2",        8, false,   4,  2670},
        {"color16-none.tif",     10,    "none bps=1",        1, false,   2,  1458},
*********/
/******* Should not test that, because JPEG only support RGB, RGBA, and all gray
        {"color256-none.tif",   10,     "jpeg=10",           8, false, 256, 4},
        {"color64-none.tif",    -1,     "jpeg",              8, false,  64, 5},
        {"color16-none.tif",    25,     "jpeg=25",           8, false,  16, 6},
        {"bitmap-lzw-pc.tif",   100,    "jpeg",              1, false,   2, 7},
*********/

        // Make sure untested configurations are indeed not supported by this libtiff
        // version. And properly generate non-fatal error message(s).

#ifndef CCITT_SUPPORT
        "bitmap-fax=3.tif",
        "bitmap-fax=4.tif",
#endif
#ifndef LZW_SUPPORT
        "cmyk8-lzw.tif",
        "lab8-lzw.tif",
        "lab16-lzw.tif",
        "rgb8-lzw-mac.tif",
        "rgb8-lzw-pc.tif",
        "rgb16-lzw-mac.tif",
        "color64-lzw-mac.tif",
        "color64-lzw-pc.tif",
        "gray8-lzw-mac.tif",
        "gray16-lzw-pc.tif",
        "bitmap-lzw-pc.tif",
        "rgb16Lzw-layersZip.tif",
#endif
#ifndef ZIP_SUPPORT
        "rgb8-zip-mac.tif",
        "rgb8-zip-pc.tif",
        "rgb16-zip-pyrm.tif",
        "gray8-zip-pc.tif",
        "gray16-zip-mac.tif",
        "bitmap-zip-pc.tif",
#endif
#ifndef JPEG_SUPPORT
        "gray8-jpeg.tif",
        "rgb8-jpeg-YCrCb.tif",
        "rgb8-jpeg-RGB.tif",
#endif
#ifndef PIXARLOG_SUPPORT
        "rgb8-pixarlog.tif",
        "rgb8-pixarlog-alpha.tif",
#endif
        0
    };

    for(const char **t = unsuportedExamples; *t; ++t)
    {
        qFatals = qWarnings = qDebugs = 0;
        QString name = QString("Examples/%1").arg(*t);

        if( ! QFileInfo(name).exists() )
            fprintf(stderr, "%d: missing file %s\n", ++error, *t);
        else
        {
            QImage img(name, "TIFF" );
            if( ! img.isNull() )
                fprintf(stderr, "%d: Should not have been able to load unsupported image %s\n", ++error, *t);
            if( qFatals || !qWarnings )
                fprintf(stderr, "%d: Failed to generate warning(s) when loading unsupported image %s\n", ++error, *t);
        }
    }

    if( error == 0 )
        fprintf(stdout, "All images checked ok (%d reads, %d write/rereads, %d unsupported)\n",
            int(sizeof readExamples/sizeof(ReadExamples))-1, 
            int(sizeof writeExamples/sizeof(WriteExamples))-1,
            int(sizeof unsuportedExamples/sizeof(char*))-1);
    else
        fprintf(stderr, "*** %d errors found. Test FAILED ***\n", error);

    return -error;
}
